import json
import aiohttpx
import backoff
import functools
from lazyops.types import BaseModel, Field, lazyproperty
from lazyops.utils import ObjectEncoder
from typing import Dict, Optional, Any, List, Type, Callable, Union

from async_openai.utils.logs import logger
from async_openai.utils.config import get_settings, get_default_headers, get_max_retries, OpenAISettings, AzureOpenAISettings
from async_openai.types.errors import fatal_exception, error_handler
from async_openai.types.resources import BaseResource, FileResource
from async_openai.types.responses import BaseResponse

__all__ = [
    'BaseRoute',
    'RESPONSE_SUCCESS_CODES',
]


"""
Route Class
"""

RESPONSE_SUCCESS_CODES = [
    200, 
    201, 
    202, 
    204
]

_retry_wrapper: Optional[Callable] = None

def get_retry_wrapper(
    max_retries: int
):
    """
    Creates the retryable wrapper
    """
    global _retry_wrapper
    if _retry_wrapper is None:
        _retry_wrapper = functools.partial(
            backoff.on_exception,
            backoff.expo, 
            exception = Exception, 
            giveup = fatal_exception
        )
    return _retry_wrapper(max_tries = max_retries + 1)


class BaseRoute(BaseModel):
    client: aiohttpx.Client
    name: str # Client Name
    # headers: Dict[str, str] = Field(default_factory = get_default_headers)
    success_codes: Optional[List[int]] = RESPONSE_SUCCESS_CODES
    
    input_model: Optional[Type[BaseResource]] = None
    response_model: Optional[Type[BaseResponse]] = None

    # Options
    timeout: Optional[int] = None
    debug_enabled: Optional[bool] = False
    on_error: Optional[Callable] = None
    ignore_errors: Optional[bool] = False
    disable_retries: Optional[bool] = None
    max_retries: Optional[int] = None
    retry_function: Optional[Callable] = None # Allow for customized retry functions

    settings: Optional[Union[OpenAISettings, AzureOpenAISettings]] = Field(default_factory = get_settings)
    is_azure: Optional[bool] = None
    azure_model_mapping: Optional[Dict[str, str]] = None

    client_callbacks: Optional[List[Callable]] = None

    @lazyproperty
    def api_resource(self):
        """
        Returns the API Resource
        """
        return ''
    
    """
    Enable or Disable Routes
    """
    @lazyproperty
    def create_enabled(self):
        return True
    
    @lazyproperty
    def create_batch_enabled(self):
        return False
    
    @lazyproperty
    def get_enabled(self):
        return False

    @lazyproperty
    def list_enabled(self):
        return False

    @lazyproperty
    def update_enabled(self):
        return False
    
    @lazyproperty
    def delete_enabled(self):
        return False


    @lazyproperty
    def download_enabled(self):
        return False
    
    @lazyproperty
    def upload_enabled(self):
        return False
    
    @lazyproperty
    def exclude_null(self) -> bool:
        return True
    

    @lazyproperty
    def usage_enabled(self):
        return False
    
    def get_resource_url(self, data: Optional[Dict[str, Any]] = None, **kwargs) -> str:
        """
        Returns the Resource URL from the Response Data
        """
        if self.is_azure is None: self.is_azure = isinstance(self.settings, AzureOpenAISettings)
        if not self.is_azure: return self.api_resource
        data = data or {}
        # logger.info(f"Data: {data}")
        base_endpoint = '/openai'
        deployment = data.get('deployment', data.get('model', kwargs.get('deployment', kwargs.get('model'))))
        if deployment is not None:
            base_endpoint += f'/deployments/{deployment}'
        base_endpoint += f'/{self.api_resource}'
        if api_version := data.get('api_version', kwargs.get('api_version', self.settings.api_version)):
            base_endpoint += f'?api-version={api_version}'
        return base_endpoint


    def encode_data(self, data: Dict[str, Any]) -> str:
        """
        Encodes the data
        """
        # if self.is_azure:
        #     _ = data.pop('model', None)
        #     _ = data.pop('deployment', None)
        return json.dumps(
            data,
            cls = ObjectEncoder
        )

    def create(
        self, 
        input_object: Optional[Type[BaseResource]] = None,
        headers: Optional[Dict[str, str]] = None,
        parse_stream: Optional[bool] = True,
        timeout: Optional[int] = None,
        **kwargs
    ):
        """
        Create a Resource

        :param input_object: Input Object to Create
        """
        if not self.create_enabled:
            raise NotImplementedError(f'Create is not enabled for {self.api_resource}')
        
        if input_object is None: 
            input_object, kwargs = self.input_model.create_resource(
                resource = self.input_model,
                **kwargs
            )
        # data = get_pyd_dict(input_object, exclude_none = self.exclude_null)
        data = input_object.dict(exclude_none = self.exclude_null)
        api_response = self._send(
            method = 'POST',
            url = self.get_resource_url(data = data, **kwargs),
            data = self.encode_data(data),
            headers = headers,
            timeout = timeout if timeout is not None else self.timeout,
            stream = input_object.get('stream'),
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(data, input_object = input_object, parse_stream = parse_stream)
    

    async def async_create(
        self, 
        input_object: Optional[Type[BaseResource]] = None,
        headers: Optional[Dict[str, str]] = None,
        parse_stream: Optional[bool] = True,
        timeout: Optional[int] = None,
        **kwargs
    ):
        """
        Create a Resource

        :param input_object: Input Object to Create
        """
        if not self.create_enabled:
            raise NotImplementedError(f'Create is not enabled for {self.api_resource}')
        
        if input_object is None:
            input_object, kwargs = self.input_model.create_resource(
                resource = self.input_model,
                **kwargs
            )
        # data = get_pyd_dict(input_object, exclude_none = self.exclude_null)
        data = input_object.dict(exclude_none = self.exclude_null)
        api_response = await self._async_send(
            method = 'POST',
            url = self.get_resource_url(data = data, **kwargs),
            data = self.encode_data(data),
            headers = headers,
            timeout = timeout if timeout is not None else self.timeout,
            stream = input_object.get('stream'),
            **kwargs
        )
        # if input_object.get('stream'):
        #     await api_response.aread()
        data = self.handle_response(api_response)
        return await self.aprepare_response(data, input_object = input_object, parse_stream = parse_stream)
    
    acreate = async_create
    
    def batch_create(
        self, 
        input_object: Optional[Type[BaseResource]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ):
        """
        Batch Create Resources

        :param input_object: Input Object to Create
        :param return_response: Return the Response Object
        """
        if not self.create_batch_enabled:
            raise NotImplementedError(f'Batch Create is not enabled for {self.api_resource}')
        
        if input_object is None:
            input_object, kwargs = self.input_model.create_resource(
                resource = self.input_model,
                **kwargs
            )

        api_resource = f'{self.api_resource}/batch'
        data = json.dumps(
            # get_pyd_dict(input_object, exclude_none = self.exclude_null),
            input_object.dict(exclude_none = self.exclude_null), 
            cls = ObjectEncoder
        )
        api_response = self._send(
            method = 'POST',
            url = self.get_resource_url(data = data),
            data = data,
            headers = headers,
            timeout = self.timeout,
            **kwargs
        )
        resp = self.handle_response(api_response)
        return self.prepare_response(resp, input_object = input_object)

    async def async_batch_create(
        self, 
        input_object: Optional[Type[BaseResource]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ):
        """
        Batch Create Resources

        :param input_object: Input Object to Create
        :param return_response: Return the Response Object
        """
        if not self.create_batch_enabled:
            raise NotImplementedError(f'Batch Create is not enabled for {self.api_resource}')
        
        if input_object is None:
            input_object, kwargs = self.input_model.create_resource(
                resource = self.input_model,
                **kwargs
            )

        api_resource = f'{self.api_resource}/batch'
        data = json.dumps(
            # get_pyd_dict(input_object, exclude_none = self.exclude_null),
            input_object.dict(exclude_none = self.exclude_null), 
            cls = ObjectEncoder
        )
        api_response = await self._async_send(  
            method = 'POST',
            url = api_resource,
            data = data,
            headers = headers,
            timeout = self.timeout,
            **kwargs
        )
        resp = self.handle_response(api_response)
        return await self.aprepare_response(resp, input_object = input_object)
    
    abatch_create = async_batch_create

    
    def retrieve(
        self, 
        resource_id: str, 
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ) -> Type[BaseResource]:
        """
        GET a Single Resource

        :param resource_id: The ID of the Resource to GET
        :param params: Optional Query Parameters
        """
        if not self.get_enabled:
            raise NotImplementedError(f'Get is not enabled for {self.api_resource}')
        
        api_resource = f'{self.api_resource}/{resource_id}'
        api_response = self._send(
            method = 'GET',
            url = api_resource, 
            params = params,
            headers = headers,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(data)
    
    async def async_retrieve(
        self,
        resource_id: str,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ) -> Type[BaseResource]:
        """
        GET a Single Resource

        :param resource_id: The ID of the Resource to GET
        :param params: Optional Query Parameters
        """
        if not self.get_enabled:
            raise NotImplementedError(f'Get is not enabled for {self.api_resource}')
        
        api_resource = f'{self.api_resource}/{resource_id}'
        api_response = await self._async_send(
            method = 'GET',
            url = api_resource, 
            params = params,
            headers = headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(data)
    
    aretrieve = async_retrieve

    def get(
        self, 
        resource_id: str, 
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ) -> Type[BaseResource]:
        """
        GET a Single Resource

        :param resource_id: The ID of the Resource to GET
        :param params: Optional Query Parameters
        """
        return self.retrieve(resource_id = resource_id, params = params, headers = headers, **kwargs)
    
    async def async_get(
        self,
        resource_id: str,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ) -> Type[BaseResource]:
        """
        GET a Single Resource

        :param resource_id: The ID of the Resource to GET
        :param params: Optional Query Parameters
        """
        return await self.async_retrieve(resource_id = resource_id, params = params, headers = headers, **kwargs)

    aget = async_get
    
    def list(
        self, 
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ) -> Dict[str, Union[List[Type[BaseResource]], Dict[str, Any]]]:
        """
        GET all available objects of Resource

        :param params: Optional Query Parameters
        
        :return: Dict[str, Union[List[Type[BaseResource]], Dict[str, Any]]]
        """
        if not self.list_enabled:
            raise NotImplementedError(f'List is not enabled for {self.api_resource}')
        
        api_response = self._send(
            method = 'GET',
            url = self.get_resource_url(data = None, **kwargs),
            params = params,
            headers = headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(data)
        #return self.prepare_index_response(data)

    async def async_list(
        self, 
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ) -> Dict[str, Union[List[Type[BaseResource]], Dict[str, Any]]]:
        """
        GET all available objects of Resource

        :param params: Optional Query Parameters

        :return: Dict[str, Union[List[Type[BaseResource]], Dict[str, Any]]]
        """
        if not self.list_enabled:
            raise NotImplementedError(f'List is not enabled for {self.api_resource}')
        
        api_response = await self._async_send(
            method = 'GET',
            url = self.get_resource_url(data = None, **kwargs),
            params = params,
            headers = headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return await self.aprepare_response(data)
    
    alist = async_list

    def get_all(
        self, 
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> Dict[str, Union[List[Type[BaseResource]], Dict[str, Any]]]:
        """
        GET all available objects of Resource

        :param params: Optional Query Parameters
        
        :return: Dict[str, Union[List[Type[BaseResource]], Dict[str, Any]]]
        """
        return self.retrieve(params = params, **kwargs)

    async def async_get_all(
        self, 
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> Dict[str, Union[List[Type[BaseResource]], Dict[str, Any]]]:
        """
        GET all available objects of Resource

        :param params: Optional Query Parameters

        :return: Dict[str, Union[List[Type[BaseResource]], Dict[str, Any]]]
        """
        return await self.async_retrieve(params = params, **kwargs)
    
    aget_all = async_get_all

    def delete(
        self, 
        resource_id: str,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ):
        """
        DELETE a Resource

        :param resource_id: The ID of the Resource to DELETE
        """
        if not self.delete_enabled:
            raise NotImplementedError(f'Delete is not enabled for {self.api_resource}')
        
        api_resource = f'{self.api_resource}/{resource_id}'
        api_response = self._send(
            method = 'DELETE',
            url = api_resource,
            headers = headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(data)
    
    async def async_delete(
        self, 
        resource_id: str,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ):
        """
        DELETE a Resource

        :param resource_id: The ID of the Resource to DELETE
        """
        if not self.delete_enabled:
            raise NotImplementedError(f'Delete is not enabled for {self.api_resource}')
        
        api_resource = f'{self.api_resource}/{resource_id}'
        api_response = await self._async_send(
            method = 'DELETE',
            url = api_resource,
            headers = headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(data)

    adelete = async_delete

    def update(
        self, 
        input_object: Optional[Type[BaseResource]] = None,
        resource_id: str = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ):
        """
        Update a Resource

        :param input_object: Input Object to Update
        :param resource_id: The ID of the Resource to Update
        """
        if not self.update_enabled:
            raise NotImplementedError(f'Update is not enabled for {self.api_resource}')
        
        if input_object is None:
            input_object, kwargs = self.input_model.create_resource(
                resource = self.input_model,
                **kwargs
            )
        
        api_resource = self.api_resource

        resource_id = resource_id or input_object.resource_id
        if resource_id is not None:
            api_resource = f'{api_resource}/{resource_id}'

        data = json.dumps(
            input_object.dict(exclude_none = self.exclude_null), 
            cls = ObjectEncoder
        )
        api_response = self._send(
            method = 'PUT',
            url = api_resource,
            data = data,
            headers = headers,

            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(data, input_object = input_object)

    async def async_update(
        self, 
        input_object: Optional[Type[BaseResource]] = None,
        resource_id: str = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ):
        """
        Update a Resource

        :param input_object: Input Object to Update
        :param resource_id: The ID of the Resource to Update
        """
        if not self.update_enabled:
            raise NotImplementedError(f'Update is not enabled for {self.api_resource}')
        
        if input_object is None: 
            input_object, kwargs = self.input_model.create_resource(
                resource = self.input_model,
                **kwargs
            )
        
        api_resource = self.api_resource
        resource_id = resource_id or input_object.resource_id
        if resource_id is not None:
            api_resource = f'{api_resource}/{resource_id}'
        data = json.dumps(
            input_object.dict(exclude_none = self.exclude_null), 
            cls = ObjectEncoder
        )
        api_response = await self._async_send(
            method = 'PUT',
            url = api_resource,
            data = data,
            headers = headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(data, input_object = input_object)

    aupdate = async_update

    """
    Extra Methods
    """

    def exists(
        self,
        resource_id: str,
        **kwargs
    ) -> bool:
        """
        See whether a Resource Exists

        :param resource_id: The ID of the Resource to Valid
        """
        try:
            return self.get(resource_id = resource_id, **kwargs)
        except Exception:
            return False
    
    async def async_exists(
        self,
        resource_id: str,
        **kwargs
    ) -> bool:
        """
        See whether a Resource Exists

        :param resource_id: The ID of the Resource to Valid
        """
        try:
            return await self.async_get(resource_id = resource_id, **kwargs)
        except Exception:
            return False
        
    aexists = async_exists
    
    def upsert(
        self,
        resource_id: str,
        input_object: Optional[Type[BaseResource]] = None,
        update_existing: bool = False, 
        overwrite_existing: bool = False,
        **kwargs
    ):
        """
        Upsert a Resource
        Validates whether the Resource exists, and if it does, updates it. 
        If it doesn't, creates it.
        If update_existing is True, it will always update the Resource
        If overwrite_existing is True, it will re-create the Resource

        :resource_id: The ID of the Resource to Upsert
        :param input_object: Input Object to Upsert
        :param update_existing (bool): Whether to update the Resource if it exists
        :overwrite_existing (bool): Whether to overwrite the Resource if it exists
        """
        resource = self.exists(resource_id = resource_id, **kwargs)
        if resource is not None:
            if update_existing:
                return self.update(input_object = input_object, identifier = resource_id, **kwargs)
            if overwrite_existing:
                self.delete(resource_id = resource_id, **kwargs)
                return self.create(input_object = input_object, **kwargs)
            return resource
        return self.create(input_object = input_object, **kwargs)
    
    async def async_upsert(
        self,
        resource_id: str,
        input_object: Optional[Type[BaseResource]] = None,
        update_existing: bool = False, 
        overwrite_existing: bool = False,
        **kwargs
    ):
        """
        Upsert a Resource
        Validates whether the Resource exists, and if it does, updates it. 
        If it doesn't, creates it.
        If update_existing is True, it will always update the Resource
        If overwrite_existing is True, it will re-create the Resource

        :resource_id: The ID of the Resource to Upsert
        :param input_object: Input Object to Upsert
        :param update_existing (bool): Whether to update the Resource if it exists
        :overwrite_existing (bool): Whether to overwrite the Resource if it exists
        """
        resource = await self.async_exists(resource_id = resource_id, **kwargs)
        if resource is not None:
            if update_existing:
                return self.async_update(input_object = input_object, identifier = resource_id, **kwargs)
            if overwrite_existing:
                await self.async_delete(resource_id = resource_id, **kwargs)
                return await self.async_create(input_object = input_object, **kwargs)
            return resource
        return await self.async_create(input_object = input_object, **kwargs)

    aupsert = async_upsert

    def upload(
        self, 
        input_object: Optional[Type[FileResource]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs,
    ):
        """
        Upload a Resource

        :param input_object: Input Object to Create
        """
        if not self.upload_enabled:
            raise NotImplementedError(f'Upload is not enabled for {self.api_resource}')
        
        if input_object is None: 
            input_object, kwargs = self.input_model.create_resource(
                resource = self.input_model,
                **kwargs
            )
        
        # headers = self.headers.copy()
        headers = headers or {}
        headers['Content-Type'] = 'multipart/form-data'

        api_resource = self.api_resource
        api_response = self._send(
            method = 'POST',
            url = api_resource,
            files = input_object.get_params(**kwargs),
            headers = headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(data, input_object = input_object)
        
    async def async_upload(
        self, 
        input_object: Optional[Type[FileResource]] = None,
        headers: Optional[Dict[str, str]] = None,
        **kwargs,
    ):
        """
        Upload a Resource

        :param input_object: Input Object to Create
        """
        if not self.upload_enabled:
            raise NotImplementedError(f'Upload is not enabled for {self.api_resource}')

        if input_object is None: 
            input_object, kwargs = self.input_model.create_resource(
                resource = self.input_model,
                **kwargs
            )
        
        # headers = self.headers.copy()
        headers = headers or {}
        headers['Content-Type'] = 'multipart/form-data'

        api_resource = self.api_resource
        api_response = await self._async_send(
            method = 'POST',
            url = api_resource,
            files = input_object.get_params(**kwargs),
            headers = headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(data, input_object = input_object)

    aupload = async_upload

    def download(
        self, 
        resource_id: str,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ):
        """
        Download a Resource

        :param resource_id: The ID of the Resource to Download
        """
        if not self.download_enabled:
            raise NotImplementedError(f'Download is not enabled for {self.api_resource}')
        
        api_resource = f'{self.api_resource}/{resource_id}/download'
        api_response = self._send(
            method = 'POST',
            url = api_resource, 
            headers = headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(data)

    async def async_download(
        self, 
        resource_id: str,
        headers: Optional[Dict[str, str]] = None,
        **kwargs
    ):
        """
        Download a Resource

        :param resource_id: The ID of the Resource to Download
        """
        if not self.download_enabled:
            raise NotImplementedError(f'Download is not enabled for {self.api_resource}')
        
        api_resource = f'{self.api_resource}/{resource_id}/download'
        api_response = await self._async_send(
            method = 'POST',
            url = api_resource, 
            headers = headers,
            timeout = self.timeout,
            **kwargs
        )
        data = self.handle_response(api_response)
        return self.prepare_response(data)

    adownload = async_download

    def prepare_response(
        self, 
        data: aiohttpx.Response,
        input_object: Optional[Type[BaseResource]] = None,
        response_object: Optional[Type[BaseResource]] = None,
        parse_stream: Optional[bool] = True,
        **kwargs
    ):
        """
        Prepare the Response Object
        
        :param data: The Response Data
        :param response_object: The Response Object
        """
        response_object = response_object or self.response_model
        if response_object:
            response = response_object.prepare_response(data, input_object = input_object, parse_stream = parse_stream)
            self.handle_callbacks(response, **kwargs)
            return response
        raise NotImplementedError('Response model not defined for this resource.')

    async def aprepare_response(
        self, 
        data: aiohttpx.Response,
        input_object: Optional[Type[BaseResource]] = None,
        response_object: Optional[Type[BaseResource]] = None,
        parse_stream: Optional[bool] = True,
        **kwargs
    ):
        """
        Prepare the Response Object
        
        :param data: The Response Data
        :param response_object: The Response Object
        """
        response_object = response_object or self.response_model
        if response_object:
            response = await response_object.aprepare_response(data, input_object = input_object, parse_stream = parse_stream)
            self.handle_callbacks(response, **kwargs)
            return response
        raise NotImplementedError('Response model not defined for this resource.')

    def handle_callbacks(
        self,
        response_object: BaseResource,
        **kwargs
    ):
        """
        Handle the Callbacks for the Response as a Background Task

        This is useful for when you want to run a background task after a response is received

        The callback should be a function that takes the response object as the first argument
        """
        if self.client_callbacks:
            from lazyops.libs.pooler import ThreadPooler
            for callback in self.client_callbacks:
                ThreadPooler.background(callback, response_object, **kwargs)
        
    def handle_response(
        self, 
        response: aiohttpx.Response,
        **kwargs
    ):
        """
        Handle the Response

        :param response: The Response
        """
        is_stream = not hasattr(response, "_content")
        if self.debug_enabled:
            if is_stream:
                logger.info(f'[{self.name} - Stream] [{response.status_code} - {response.request.url}] headers: {response.headers}')
            else:
                logger.info(f'[{self.name} - {response.status_code} - {response.request.url}] headers: {response.headers}, body: {response.text[:250]}')
        
        if response.status_code in self.success_codes:
            if is_stream: return response
            return response if response.content else None
        
        if self.ignore_errors: return None
        raise error_handler(
            response = response,
            data = response.content,
            **kwargs
        )
    
    def _send(
        self,
        method: str,
        url: str,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        stream: Optional[bool] = False,
        **kwargs
    ) -> aiohttpx.Response:
        """
        Handle Sending Requests
        """
        if timeout is None: timeout = self.timeout
        if self.debug_enabled: logger.info(f'[{self.name} - {method} - {url}] headers: {headers}, params: {params}, data: {data}')
        request = self.client.build_request(
            method = method,
            url = url,
            params = params,
            data = data,
            headers = headers,
            timeout = timeout,
            **kwargs
        )
        request_func = self.client.send
        if self.retry_function is not None:
            request_func = self.retry_function(request_func)
        elif not self.disable_retries:
            if max_retries is None: max_retries = get_max_retries()
            request_func = get_retry_wrapper(max_retries=max_retries)(request_func)
        return request_func(request, stream = stream)

    
    async def _async_send(
        self,
        method: str,
        url: str,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        stream: Optional[bool] = False,
        **kwargs
    ) -> aiohttpx.Response:
        """
        Handle Sending Requests
        """
        
        if timeout is None: timeout = self.timeout
        # if self.debug_enabled: logger.info(f'[{self.name} - {method} - {url}] headers: {headers}, params: {params}, data: {data}')
        request = await self.client.async_build_request(
            method = method,
            url = url,
            params = params,
            data = data,
            headers = headers,
            timeout = timeout,
            **kwargs
        )
        if self.debug_enabled: logger.info(f'[{self.name} - {method} - {url}] headers: {request.headers}, params: {params}, data: {data}')
        request_func = self.client.async_send
        if self.retry_function is not None:
            request_func = self.retry_function(request_func)
        elif not self.disable_retries:
            if max_retries is None: max_retries = get_max_retries()
            request_func = get_retry_wrapper(max_retries = max_retries)(request_func)
        return await request_func(request, stream = stream)
    
    def prepare_index_response(
        self, 
        data: Dict[str, Any],
        response_object: Optional[Type[BaseResource]] = None,
        **kwargs
    ):
        """
        Prepare the Response Object for Index Requests

        :param data: The Response Data
        :param response_object: The Response Object
        """
        collection = [
            self.prepare_response(el, response_object = response_object) for el in data[self.api_resource]
        ]
        return {
            self.api_resource: collection,
            'meta': data['meta']
        }
    




"""
Client Metaclass
"""

from __future__ import annotations

import pathlib

from typing import Optional, List, Callable, Dict, Union, overload, TYPE_CHECKING

from async_openai.schemas import *
from async_openai.types.options import ApiType
from async_openai.types.context import ModelContextHandler
from async_openai.utils.config import get_settings, OpenAISettings
from async_openai.utils.logs import logger

if TYPE_CHECKING:
    from async_openai.client import OpenAIClient


class RotatingClients:
    """
    Manages a set of clients that can be rotated.    
    """
    def __init__(self, prioritize: Optional[str] = None, settings: Optional[OpenAISettings] = None, azure_model_mapping: Optional[Dict[str, str]] = None):
        self.settings = settings or get_settings()
        self.clients: Dict[str, 'OpenAIClient'] = {}
        self.rotate_index: int = 0
        self.rotate_client_names: List[str] = []
        self.azure_model_mapping: Dict[str, str] = azure_model_mapping

        assert prioritize in [None, 'azure', 'openai'], f'Invalid `prioritize` value: {prioritize}'
        self.prioritize: Optional[str] = prioritize
    
    @property
    def client_names(self) -> List[str]:
        """
        Returns the list of client names.
        """
        return list(self.clients.keys())
    
    def run_client_init(self):
        """
        Initializes the Client. 

        Can be subclassed to provide custom initialization.
        """
        self.init_api_client()
        if self.settings.has_valid_azure:
            self.init_api_client(client_name = 'az', is_azure = True, set_as_default = self.prioritize == 'azure', set_as_current = self.prioritize == 'azure')

    @property
    def api(self) -> 'OpenAIClient':
        """
        Returns the inherited OpenAI client.
        """
        if not self.clients: 
            self.run_client_init()
        if not self.rotate_client_names:
            return self.clients[self.client_names[self.rotate_index]]
        return self.clients[self.rotate_client_names[self.rotate_index]]
    
    def increase_rotate_index(self):
        """
        Increases the rotate index
        """
        if self.rotate_index >= len(self.clients) - 1:
            self.rotate_index = 0
        else:
            self.rotate_index += 1

    
    def rotate_client(self, index: Optional[int] = None, require_azure: Optional[bool] = None, verbose: Optional[bool] = False):
        """
        Rotates the clients
        """
        if index is not None:
            self.rotate_index = index
            return
        self.increase_rotate_index()
        if require_azure:
            while not self.api.is_azure:
                self.increase_rotate_index()
        if verbose:
            logger.info(f'Rotated Client: {self.api.name} (Azure: {self.api.is_azure} - {self.api.api_version}) [{self.rotate_index+1}/{len(self.clients)}]')
    
    def set_client(self, client_name: Optional[str] = None, verbose: Optional[bool] = False):
        """
        Sets the client
        """
        if client_name is None:
            raise ValueError('`client_name` is required.')
        if client_name not in self.clients:
            raise ValueError(f'Client `{client_name}` does not exist.')
        self.rotate_index = self.client_names.index(client_name)
        if verbose:
            logger.info(f'Set Client: {self.api.name} (Azure: {self.api.is_azure} - {self.api.api_version})) [{self.rotate_index+1}/{len(self.clients)}]')

    def current_client_info(self, verbose: Optional[bool] = False) -> Dict[str, Union[str, int]]:
        """
        Returns the current client info
        """
        data = {
            'name': self.api.name,
            'is_azure': self.api.is_azure,
            'api_version': self.api.api_version,
            'index': self.rotate_index,
            'total': len(self.clients),
        }
        if verbose:
            logger.info(f'Current Client: {self.api.name} (Azure: {self.api.is_azure} - {self.api.api_version}) [{self.rotate_index+1}/{len(self.clients)}]')
        return data


    def configure_client(self, client_name: Optional[str] = None, priority: Optional[int] = None, **kwargs):
        """
        Configure a new client
        """
        client_name = client_name or 'default'
        if client_name not in self.clients:
            raise ValueError(f'Client `{client_name}` does not exist.')
        self.clients[client_name].reset(**kwargs)
        if priority is not None:
            if client_name in self.rotate_client_names:
                self.rotate_client_names.remove(client_name)
            self.rotate_client_names.insert(priority, client_name)

    def init_api_client(
        self, 
        client_name: Optional[str] = None, 
        set_as_default: Optional[bool] = False, 
        is_azure: Optional[bool] = None,
        priority: Optional[int] = None,
        set_as_current: Optional[bool] = False,
        **kwargs
    ) -> 'OpenAIClient':
        """
        Creates a new OpenAI client.
        """
        client_name = client_name or 'default'
        if client_name in self.clients:
            return self.clients[client_name]

        from async_openai.client import OpenAIClient
        if is_azure is None and \
                (
                'az' in client_name and self.settings.has_valid_azure
            ):
            is_azure = True
        client = OpenAIClient(
            name = client_name,
            settings = self.settings,
            is_azure = is_azure,
            azure_model_mapping = self.azure_model_mapping,
            **kwargs
        )
        self.clients[client_name] = client
        if set_as_default:
            self.rotate_client_names.insert(0, client_name)
        elif priority is not None:
            if client_name in self.rotate_client_names:
                self.rotate_client_names.remove(client_name)
            self.rotate_client_names.insert(priority, client_name)
        elif self.prioritize:
            if (
                self.prioritize == 'azure'
                and is_azure
                or self.prioritize != 'azure'
                and self.prioritize == 'openai'
                and not is_azure
            ):
                self.rotate_client_names.insert(0, client_name)
            elif self.prioritize in ['azure', 'openai']:
                self.rotate_client_names.append(client_name)
        if set_as_current:
            self.rotate_index = self.rotate_client_names.index(client_name)
        return client
    
    def get_api_client(self, client_name: Optional[str] = None, require_azure: Optional[bool] = None, **kwargs) -> 'OpenAIClient':
        """
        Initializes a new OpenAI client or Returns an existing one.
        """
        if not client_name and not self.clients:
            client_name = 'default'
        if client_name and client_name not in self.clients:
            self.clients[client_name] = self.init_api_client(client_name = client_name, **kwargs)

        if not client_name and require_azure:
            while not self.api.is_azure:
                self.increase_rotate_index()
            return self.api
        return self.clients[client_name] if client_name else self.api

    def __getitem__(self, key: Union[str, int]) -> 'OpenAIClient':
        """
        Returns a client by name.
        """
        if isinstance(key, int):
            key = self.rotate_client_names[key] if self.rotate_client_names else self.client_names[key]
        return self.clients[key]

# Model Mapping for Azure
DefaultModelMapping = {
    'gpt-3.5-turbo': 'gpt-35-turbo',
    'gpt-3.5-turbo-16k': 'gpt-35-turbo-16k',
    'gpt-3.5-turbo-instruct': 'gpt-35-turbo-instruct',
    'gpt-3.5-turbo-0301': 'gpt-35-turbo-0301',
    'gpt-3.5-turbo-0613': 'gpt-35-turbo-0613',
    'gpt-3.5-turbo-1106': 'gpt-35-turbo-1106',
}
    
class OpenAIMetaClass(type):
    # api_key: Optional[str] = None
    # url: Optional[str] = None
    # scheme: Optional[str] = None
    # host: Optional[str] = None
    # port: Optional[int] = None

    # api_key: Optional[str] = None
    # api_path: Optional[str] = None
    # api_type: Optional[ApiType] = None
    # api_version: Optional[str] = None
    # api_key_path: Optional[pathlib.Path] = None
    
    # organization: Optional[str] = None
    # proxies: Optional[Union[str, Dict]] = None
    # app_info: Optional[Dict[str, str]] = None
    
    # debug_enabled: Optional[bool] = None
    # ignore_errors: Optional[bool] = None
    # max_retries: Optional[int] = None
    # timeout: Optional[int] = None

    # headers: Optional[Dict] = None

    on_error: Optional[Callable] = None
    prioritize: Optional[str] = None
    enable_rotating_clients: Optional[bool] = False
    azure_model_mapping: Optional[Dict[str, str]] = DefaultModelMapping

    _api: Optional['OpenAIClient'] = None
    _apis: Optional['RotatingClients'] = None
    _clients: Optional[Dict[str, 'OpenAIClient']] = {}
    _settings: Optional[OpenAISettings] = None

    """
    The Global Meta Class for OpenAI API.
    """

    @property
    def settings(cls) -> OpenAISettings:
        """
        Returns the global settings for the OpenAI API.
        """
        if cls._settings is None:
            cls._settings = get_settings()
        return cls._settings
    
    # Changing the behavior to become proxied through settings

    @property
    def api_key(cls) -> Optional[str]:
        """
        Returns the global API Key.
        """
        return cls.settings.api_key
    
    @property
    def url(cls) -> Optional[str]:
        """
        Returns the global URL.
        """
        return cls.settings.url
    
    @property
    def scheme(cls) -> Optional[str]:
        """
        Returns the global Scheme.
        """
        return cls.settings.scheme
    
    @property
    def host(cls) -> Optional[str]:
        """
        Returns the global Host.
        """
        return cls.settings.host
    
    @property
    def port(cls) -> Optional[int]:
        """
        Returns the global Port.
        """
        return cls.settings.port
    
    @property
    def api_base(cls) -> Optional[str]:
        """
        Returns the global API Base.
        """
        return cls.settings.api_base

    @property
    def api_path(cls) -> Optional[str]:
        """
        Returns the global API Path.
        """
        return cls.settings.api_path
    
    @property
    def api_type(cls) -> Optional[ApiType]:
        """
        Returns the global API Type.
        """
        return cls.settings.api_type
    
    @property
    def api_version(cls) -> Optional[str]:
        """
        Returns the global API Version.
        """
        return cls.settings.api_version
    
    @property
    def api_key_path(cls) -> Optional[pathlib.Path]:
        """
        Returns the global API Key Path.
        """
        return cls.settings.api_key_path
    
    @property
    def organization(cls) -> Optional[str]:
        """
        Returns the global Organization.
        """
        return cls.settings.organization
    
    @property
    def proxies(cls) -> Optional[Union[str, Dict]]:
        """
        Returns the global Proxies.
        """
        return cls.settings.proxies
    
    @property
    def timeout(cls) -> Optional[int]:
        """
        Returns the global Timeout.
        """
        return cls.settings.timeout
    
    @property
    def max_retries(cls) -> Optional[int]:
        """
        Returns the global Max Retries.
        """
        return cls.settings.max_retries
    
    @property
    def app_info(cls) -> Optional[Dict[str, str]]:
        """
        Returns the global App Info.
        """
        return cls.settings.app_info
    
    @property
    def debug_enabled(cls) -> Optional[bool]:
        """
        Returns the global Debug Enabled.
        """
        return cls.settings.debug_enabled
    
    @property
    def ignore_errors(cls) -> Optional[bool]:
        """
        Returns the global Ignore Errors.
        """
        return cls.settings.ignore_errors
    
    @property
    def timeout(cls) -> Optional[int]:
        """
        Returns the global Timeout.
        """
        return cls.settings.timeout


    @overload
    def configure(
        cls, 
        api_key: Optional[str] = None,
        url: Optional[str] = None,
        scheme: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        
        api_base: Optional[str] = None,
        api_path: Optional[str] = None,
        api_type: Optional[ApiType] = None,
        api_version: Optional[str] = None,
        api_key_path: Optional[pathlib.Path] = None,

        organization: Optional[str] = None,
        proxies: Optional[Union[str, Dict]] = None,
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        app_info: Optional[Dict[str, str]] = None,
        debug_enabled: Optional[bool] = None,
        ignore_errors: Optional[bool] = None,
        disable_retries: Optional[bool] = None,
        max_connections: Optional[int] = None,
        max_keepalive_connections: Optional[int] = None,
        keepalive_expiry: Optional[int] = None,
        custom_headers: Optional[Dict[str, str]] = None,

        on_error: Optional[Callable] = None,
        reset: Optional[bool] = None,
        prioritize: Optional[str] = None,
        enable_rotating_clients: Optional[bool] = None,
        azure_model_mapping: Optional[Dict[str, str]] = None,
        **kwargs
    ):
        """
        Configure the global OpenAI client.
        
        :param url: The OpenAI API URL              | Env: [`OPENAI_API_URL`]
        :param scheme: The OpenAI API Scheme        | Env: [`OPENAI_API_SCHEME`]
        :param host: The OpenAI API Host            | Env: [`OPENAI_API_HOST`]
        :param port: The OpenAI API Port            | Env: [`OPENAI_API_PORT`]
        :param api_base: The OpenAI API Base        | Env: [`OPENAI_API_BASE`]
        :param api_key: The OpenAI API Key          | Env: [`OPENAI_API_KEY`]
        :param api_path: The OpenAI API Path        | Env: [`OPENAI_API_PATH`]
        :param api_type: The OpenAI API Type        | Env: [`OPENAI_API_TYPE`]
        :param api_version: The OpenAI API Version  | Env: [`OPENAI_API_VERSION`]
        :param api_key_path: The API Key Path       | Env: [`OPENAI_API_KEY_PATH`]
        :param organization: Organization           | Env: [`OPENAI_ORGANIZATION`]
        :param proxies: The OpenAI Proxies          | Env: [`OPENAI_PROXIES`]
        :param timeout: Timeout in Seconds          | Env: [`OPENAI_TIMEOUT`]
        :param max_retries: The OpenAI Max Retries  | Env: [`OPENAI_MAX_RETRIES`]
        :param ignore_errors: Ignore Errors         | Env: [`OPENAI_IGNORE_ERRORS`]
        :param disable_retries: Disable Retries     | Env: [`OPENAI_DISABLE_RETRIES`]
        :param max_connections: Max Connections     | Env: [`OPENAI_MAX_CONNECTIONS`]
        :param max_keepalive_connections: Max Keepalive Connections | Env: [`OPENAI_MAX_KEEPALIVE_CONNECTIONS`]
        :param keepalive_expiry: Keepalive Expiry   | Env: [`OPENAI_KEEPALIVE_EXPIRY`]
        :param custom_headers: Custom Headers       | Env: [`OPENAI_CUSTOM_HEADERS`]
        
        :param on_error: On Error Callback          
        :param kwargs: Additional Keyword Arguments
        """
        ...
    
    @overload
    def configure(
        azure_api_key: Optional[str] = None,
        azure_url: Optional[str] = None,
        azure_scheme: Optional[str] = None,
        azure_host: Optional[str] = None,
        azure_port: Optional[int] = None,
        
        azure_api_base: Optional[str] = None,
        azure_api_path: Optional[str] = None,
        azure_api_type: Optional[ApiType] = None,
        azure_api_version: Optional[str] = None,
        azure_api_key_path: Optional[pathlib.Path] = None,

        azure_organization: Optional[str] = None,
        azure_proxies: Optional[Union[str, Dict]] = None,
        azure_timeout: Optional[int] = None,
        azure_max_retries: Optional[int] = None,
        azure_app_info: Optional[Dict[str, str]] = None,
        azure_debug_enabled: Optional[bool] = None,
        azure_ignore_errors: Optional[bool] = None,
        azure_max_connections: Optional[int] = None,
        azure_max_keepalive_connections: Optional[int] = None,
        azure_keepalive_expiry: Optional[int] = None,
        azure_custom_headers: Optional[Dict[str, str]] = None,

        on_error: Optional[Callable] = None,
        reset: Optional[bool] = None,
        prioritize: Optional[str] = None,
        enable_rotating_clients: Optional[bool] = None,
        azure_model_mapping: Optional[Dict[str, str]] = None,
        **kwargs
    ):
        """
        Configure the global OpenAI client for Azure

        :param azure_url: The OpenAI API URL              | Env: [`AZURE_OPENAI_API_URL`]
        :param azure_scheme: The OpenAI API Scheme        | Env: [`AZURE_OPENAI_API_SCHEME`]
        :param azure_host: The OpenAI API Host            | Env: [`AZURE_OPENAI_API_HOST`]
        :param azure_port: The OpenAI API Port            | Env: [`AZURE_OPENAI_API_PORT`]
        :param azure_api_key: The OpenAI API Key          | Env: [`AZURE_OPENAI_API_KEY`]
        :param azure_api_base: The OpenAI API Base        | Env: [`AZURE_OPENAI_API_BASE`]
        :param azure_api_path: The OpenAI API Path        | Env: [`AZURE_OPENAI_API_PATH`]
        :param azure_api_type: The OpenAI API Type        | Env: [`AZURE_OPENAI_API_TYPE`]
        :param azure_api_version: The OpenAI API Version  | Env: [`AZURE_OPENAI_API_VERSION`]
        :param azure_api_key_path: The API Key Path       | Env: [`AZURE_OPENAI_API_KEY_PATH`]
        :param azure_organization: Organization           | Env: [`AZURE_OPENAI_ORGANIZATION`]
        :param azure_proxies: The OpenAI Proxies          | Env: [`AZURE_OPENAI_PROXIES`]
        :param azure_timeout: Timeout in Seconds          | Env: [`AZURE_OPENAI_TIMEOUT`]
        :param azure_max_retries: The OpenAI Max Retries  | Env: [`AZURE_OPENAI_MAX_RETRIES`]
        :param kwargs: Additional Keyword Arguments
        """
        ...

    # @overload
    def configure(
        cls, 
        on_error: Optional[Callable] = None,
        prioritize: Optional[str] = None,
        enable_rotating_clients: Optional[bool] = None,
        azure_model_mapping: Optional[Dict[str, str]] = None,
        # reset: Optional[bool] = None,
        **kwargs
    ):
        """
        Configure the global OpenAI client.
        """
        if on_error is not None: cls.on_error = on_error
        if prioritize is not None: cls.prioritize = prioritize
        if enable_rotating_clients is not None: cls.enable_rotating_clients = enable_rotating_clients
        if azure_model_mapping is not None: 
            cls.azure_model_mapping = azure_model_mapping
            for key, val in azure_model_mapping.items():
                ModelContextHandler.add_model(key, val)
        cls.settings.configure(**kwargs)
    
    def configure_client(
        cls,
        client_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Configure a specific client.
        """
        if cls.enable_rotating_clients:
            return cls.apis.configure_client(client_name = client_name, **kwargs)
        client_name = client_name or 'default'
        if client_name not in cls._clients:
            raise ValueError(f'Client `{client_name}` does not exist.')
        cls._clients[client_name].reset(**kwargs)

    def get_api_client(
        cls,
        client_name: Optional[str] = None,
        **kwargs,
    ) -> 'OpenAIClient':
        """
        Initializes a new OpenAI client or Returns an existing one.
        """
        if cls.enable_rotating_clients:
            return cls.apis.get_api_client(client_name = client_name, **kwargs)
        client_name = client_name or 'default'
        if client_name not in cls._clients:
            cls._clients[client_name] = cls.init_api_client(client_name = client_name, **kwargs)
        return cls._clients[client_name]
        
    def init_api_client(
        cls, 
        client_name: Optional[str] = None,
        set_as_default: Optional[bool] = False,
        is_azure: Optional[bool] = None,
        **kwargs
    ) -> 'OpenAIClient':
        """
        Creates a new OpenAI client.
        """
        if cls.enable_rotating_clients:
            return cls.apis.init_api_client(client_name = client_name, set_as_default = set_as_default, is_azure = is_azure, **kwargs)
        client_name = client_name or 'default'
        if client_name in cls._clients:
            return cls._clients[client_name]
        
        from async_openai.client import OpenAIClient
        if is_azure is None and \
            (
                # (client_name == 'default' or 'az' in client_name) and 
                'az' in client_name and cls.settings.has_valid_azure
            ):
            is_azure = True
        client = OpenAIClient(
            name = client_name,
            settings = cls.settings,
            is_azure = is_azure,
            azure_model_mapping = cls.azure_model_mapping,
            **kwargs
        )
        cls._clients[client_name] = client
        if set_as_default or not cls._api:
            cls._api = client
        return client
    
    def rotate_client(cls, index: Optional[int] = None, verbose: Optional[bool] = False, **kwargs):
        """
        Rotates the clients
        """
        if not cls.enable_rotating_clients:
            raise ValueError('Rotating Clients is not enabled.')
        cls.apis.rotate_client(index = index, verbose = verbose, **kwargs)
    
    def set_client(cls, client_name: Optional[str] = None, verbose: Optional[bool] = False):
        """
        Sets the client
        """
        if cls.enable_rotating_clients:
            cls.apis.set_client(client_name = client_name, verbose = verbose)
        else:
            cls._api = cls._clients[client_name]
            if verbose:
                logger.info(f'Set Client: {cls.api.name} ({cls.api.is_azure})')
    
    def get_current_client_info(cls, verbose: Optional[bool] = False) -> Dict[str, Union[str, int]]:
        """
        Returns the current client info
        """
        if cls.enable_rotating_clients:
            return cls.apis.current_client_info(verbose = verbose)
        data = {
            'name': cls.api.name,
            'is_azure': cls.api.is_azure,
            'api_version': cls.api.api_version,
        }
        if verbose:
            logger.info(f'Current Client: {cls.api.name} (Azure: {cls.api.is_azure} - {cls.api.api_version})')
        return data

    
    @property
    def apis(cls) -> RotatingClients:
        """
        Returns the global Rotating Clients.
        """
        if cls._apis is None:
            cls._apis = RotatingClients(prioritize=cls.prioritize, settings=cls.settings, azure_model_mapping=cls.azure_model_mapping)
        return cls._apis
    
    @property
    def api(cls) -> 'OpenAIClient':
        """
        Returns the inherited OpenAI client.
        """
        if cls.enable_rotating_clients: return cls.apis.api
        if cls._api is None:
            cls.init_api_client()
        return cls._api
    
    """
    API Routes
    """

    @property
    def completions(cls) -> CompletionRoute:
        """
        Returns the `CompletionRoute` class for interacting with `Completions`.
        
        Doc: `https://beta.openai.com/docs/api-reference/completions`
        """
        return cls.api.completions
    
    @property
    def Completions(cls) -> CompletionRoute:
        """
        Returns the `CompletionRoute` class for interacting with `Completions`.
        
        Doc: `https://beta.openai.com/docs/api-reference/completions`
        """
        return cls.api.completions
    

    @property
    def chat(cls) -> ChatRoute:
        """
        Returns the `ChatRoute` class for interacting with `Chat`.
        
        Doc: `https://beta.openai.com/docs/api-reference/chat`
        """
        return cls.api.chat
    
    @property
    def Chat(cls) -> ChatRoute:
        """
        Returns the `ChatRoute` class for interacting with `Chat`.
        
        Doc: `https://beta.openai.com/docs/api-reference/chat`
        """
        return cls.api.chat
    
    @property
    def edits(cls) -> EditRoute:
        """
        Returns the `EditRoute` class for interacting with `Edits`.
        
        Doc: `https://beta.openai.com/docs/api-reference/edits`
        """
        return cls.api.edits
    
    @property
    def embeddings(cls) -> EmbeddingRoute:
        """
        Returns the `EmbeddingRoute` class for interacting with `Embeddings`.
        
        Doc: `https://beta.openai.com/docs/api-reference/embeddings`
        """
        return cls.api.embeddings
    
    @property
    def images(cls) -> ImageRoute:
        """
        Returns the `ImageRoute` class for interacting with `Images`.
        
        Doc: `https://beta.openai.com/docs/api-reference/images`
        """
        return cls.api.images
    
    @property
    def models(cls) -> ModelRoute:
        """
        Returns the `ModelRoute` class for interacting with `models`.
        
        Doc: `https://beta.openai.com/docs/api-reference/models`
        """
        return cls.api.models



    """
    Context Managers
    """

    async def async_close(cls):
        """
        Closes the OpenAI API Client.
        """
        for client in cls._clients.values():
            await client.async_close()

    
    def close(cls):
        """
        Closes the OpenAI API Client.
        """
        for client in cls._clients.values():
            client.close()
    
    def __enter__(cls):
        return cls
    
    def __exit__(cls, exc_type, exc_value, traceback):
        cls.close()
    
    async def __aenter__(cls):
        return cls
    
    async def __aexit__(cls, exc_type, exc_value, traceback):
        await cls.async_close()

    def __getitem__(cls, key: Union[int, str]) -> 'OpenAIClient':
        """
        Returns the OpenAI API Client.
        """
        if cls.enable_rotating_clients:
            return cls.apis[key]
        if isinstance(key, int):
            key = cls.client_names[key]
        return cls._clients[key]
    
    @property
    def client_names(cls) -> List[str]:
        """
        Returns the list of client names.
        """
        return list(cls._clients.keys())


    """
    Auto Rotating Functions
    """

    def chat_create(
        cls, 
        input_object: Optional[ChatObject] = None,
        parse_stream: Optional[bool] = True,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        verbose: Optional[bool] = False,
        **kwargs
    ) -> ChatResponse:
        """
        Creates a chat response for the provided prompt and parameters

        Usage:

        ```python
        >>> result = OpenAI.chat_create(
        >>>    messages = [{'content': 'say this is a test'}],
        >>>    max_tokens = 4,
        >>>    stream = True
        >>> )
        ```

        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API 
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `gpt-3.5-turbo`
        
        :messages: The messages to generate chat completions for, in the chat format.

        :max_tokens (optional): The maximum number of tokens to generate in the completion.
        The token count of your prompt plus `max_tokens` cannot exceed the model's context length. 
        Most models have a context length of 2048 tokens (except for the newest models, which 
        support 4096 / 8182 / 32,768). If max_tokens is not provided, the model will use the maximum number of tokens
        Default: None

        :temperature (optional): What sampling temperature to use. Higher values means 
        the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) 
        for ones with a well-defined answer. We generally recommend altering this or `top_p` but not both.
        Default: `1.0`

        :top_p (optional): An alternative to sampling with `temperature`, called nucleus 
        sampling, where the model considers the results of the tokens with `top_p` probability mass. 
        So `0.1` means only  the tokens comprising the top 10% probability mass are considered.
        We generally recommend altering this or `temperature` but not both
        Default: `1.0`

        :n (optional): How many completions to generate for each prompt.
        Note: Because this parameter generates many completions, it can quickly 
        consume your token quota. Use carefully and ensure that you have reasonable 
        settings for `max_tokens` and stop.
        Default: `1`

        :stream (optional): CURRENTLY NOT SUPPORTED
        Whether to stream back partial progress. 
        If set, tokens will be sent as data-only server-sent events as they become 
        available, with the stream terminated by a `data: [DONE]` message. This is 
        handled automatically by the Client and enables faster response processing.
        Default: `False`

        :logprobs (optional): Include the log probabilities on the `logprobs` 
        most likely tokens, as well the chosen tokens. For example, if `logprobs` is 5, 
        the API will return a list of the 5 most likely tokens. The API will always 
        return the logprob of the sampled token, so there may be up to `logprobs+1` 
        elements in the response. The maximum value for `logprobs` is 5.
        Default: `None`

        :stop (optional): Up to 4 sequences where the API will stop generating 
        further tokens. The returned text will not contain the stop sequence.
        Default: `None`

        :presence_penalty (optional): Number between `-2.0` and `2.0`. Positive values 
        penalize new tokens based on whether they appear in the text so far, increasing the 
        model's likelihood to talk about new topics
        Default: `0.0`

        :frequency_penalty (optional): Number between `-2.0` and `2.0`. Positive 
        values penalize new tokens based on their existing frequency in the text so 
        far, decreasing the model's likelihood to repeat the same line verbatim.
        Default: `0.0`

        :logit_bias (optional): Modify the likelihood of specified tokens appearing in the completion.
        Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated 
        bias value from -100 to 100. You can use this tokenizer tool (which works for both GPT-2 and GPT-3) to 
        convert text to token IDs. Mathematically, the bias is added to the logits generated by the model prior 
        to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase 
        likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the 
        relevant token.
        As an example, you can pass `{"50256": -100}` to prevent the `<|endoftext|>` token from being generated.
        Default: `None`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.
        Default: `None`

        :functions (optional): A list of dictionaries representing the functions to call
        
        :function_call (optional): The name of the function to call. Default: `auto` if functions are provided

        :auto_retry (optional): Whether to automatically retry the request if it fails due to a rate limit error.

        :auto_retry_limit (optional): The maximum number of times to retry the request if it fails due to a rate limit error.

        Returns: `ChatResponse`
        """
    
        try:
            return cls.api.chat.create(input_object = input_object, parse_stream = parse_stream, auto_retry = auto_retry, auto_retry_limit = auto_retry_limit, **kwargs)

        except Exception as e:
            if not cls.enable_rotating_clients: raise e
            cls.rotate_client(verbose=verbose)
            return cls.chat_create(input_object = input_object, parse_stream = parse_stream, auto_retry = auto_retry, auto_retry_limit = auto_retry_limit, **kwargs)
    

    async def async_chat_create(
        cls, 
        input_object: Optional[ChatObject] = None,
        parse_stream: Optional[bool] = True,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        verbose: Optional[bool] = False,
        **kwargs
    ) -> ChatResponse:
        """
        Creates a chat response for the provided prompt and parameters

        Usage:

        ```python
        >>> result = await OpenAI.async_chat_create(
        >>>    messages = [{'content': 'say this is a test'}],
        >>>    max_tokens = 4,
        >>>    stream = True
        >>> )
        ```

        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API 
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `gpt-3.5-turbo`
        
        :messages: The messages to generate chat completions for, in the chat format.

        :max_tokens (optional): The maximum number of tokens to generate in the completion.
        The token count of your prompt plus `max_tokens` cannot exceed the model's context length. 
        Most models have a context length of 2048 tokens (except for the newest models, which 
        support 4096).
        Default: `16`

        :temperature (optional): What sampling temperature to use. Higher values means 
        the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) 
        for ones with a well-defined answer. We generally recommend altering this or `top_p` but not both.
        Default: `1.0`

        :top_p (optional): An alternative to sampling with `temperature`, called nucleus 
        sampling, where the model considers the results of the tokens with `top_p` probability mass. 
        So `0.1` means only  the tokens comprising the top 10% probability mass are considered.
        We generally recommend altering this or `temperature` but not both
        Default: `1.0`

        :n (optional): How many completions to generate for each prompt.
        Note: Because this parameter generates many completions, it can quickly 
        consume your token quota. Use carefully and ensure that you have reasonable 
        settings for `max_tokens` and stop.
        Default: `1`

        :stream (optional): CURRENTLY NOT SUPPORTED
        Whether to stream back partial progress. 
        If set, tokens will be sent as data-only server-sent events as they become 
        available, with the stream terminated by a `data: [DONE]` message. This is 
        handled automatically by the Client and enables faster response processing.
        Default: `False`

        :logprobs (optional): Include the log probabilities on the `logprobs` 
        most likely tokens, as well the chosen tokens. For example, if `logprobs` is 5, 
        the API will return a list of the 5 most likely tokens. The API will always 
        return the logprob of the sampled token, so there may be up to `logprobs+1` 
        elements in the response. The maximum value for `logprobs` is 5.
        Default: `None`

        :stop (optional): Up to 4 sequences where the API will stop generating 
        further tokens. The returned text will not contain the stop sequence.
        Default: `None`

        :presence_penalty (optional): Number between `-2.0` and `2.0`. Positive values 
        penalize new tokens based on whether they appear in the text so far, increasing the 
        model's likelihood to talk about new topics
        Default: `0.0`

        :frequency_penalty (optional): Number between `-2.0` and `2.0`. Positive 
        values penalize new tokens based on their existing frequency in the text so 
        far, decreasing the model's likelihood to repeat the same line verbatim.
        Default: `0.0`

        :logit_bias (optional): Modify the likelihood of specified tokens appearing in the completion.
        Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated 
        bias value from -100 to 100. You can use this tokenizer tool (which works for both GPT-2 and GPT-3) to 
        convert text to token IDs. Mathematically, the bias is added to the logits generated by the model prior 
        to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase 
        likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the 
        relevant token.
        As an example, you can pass `{"50256": -100}` to prevent the `<|endoftext|>` token from being generated.
        Default: `None`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.

        :functions (optional): A list of dictionaries representing the functions to call
        
        :function_call (optional): The name of the function to call. Default: `auto` if functions are provided

        :auto_retry (optional): Whether to automatically retry the request if it fails due to a rate limit error.

        :auto_retry_limit (optional): The maximum number of times to retry the request if it fails due to a rate limit error.

        Default: `None`

        Returns: `ChatResponse`
        """

        try:
            return await cls.api.chat.async_create(input_object = input_object, parse_stream = parse_stream, auto_retry = auto_retry, auto_retry_limit = auto_retry_limit, **kwargs)

        except Exception as e:
            if not cls.enable_rotating_clients: raise e
            cls.rotate_client(verbose=verbose)
            return await cls.async_chat_create(input_object = input_object, parse_stream = parse_stream, auto_retry = auto_retry, auto_retry_limit = auto_retry_limit, **kwargs)
    

    



import re
import json
from typing import Optional, Callable, Dict, Union, List, Any
from .logs import logger
from .fixjson import fix_json


_json_pattern = re.compile(r"({[^}]*$|{.*})", flags=re.DOTALL)

def build_stack(json_str: str):
    stack = []
    fixed_str = ""
    open_quotes = False

    # a flag indicating whether we've seen a comma or colon most recently
    # since last opening/closing a dict or list
    last_seen_comma_or_colon = None

    for i, char in enumerate(json_str):
        if not open_quotes:
            # opening a new nested
            if char in "{[":
                stack.append(char)
                last_seen_comma_or_colon = None
            # closing a nested
            elif char in "}]":
                stack.pop()
                last_seen_comma_or_colon = None
            if char in ",:":
                last_seen_comma_or_colon = char
        # opening or closing a string, only it's not escaped
        if char == '"' and i > 0 and json_str[i - 1] != "\\":
            open_quotes = not open_quotes

        fixed_str += char

    return (stack, fixed_str, open_quotes, last_seen_comma_or_colon)



def is_truncated(json_str: str):
    """
    Check if the json string is truncated by checking if the number of opening
    brackets is greater than the number of closing brackets.
    """
    stack, _, _, _ = build_stack(json_str)
    return len(stack) > 0


def find_json_response(full_response: str, verbose: Optional[bool] = False):
    """
    Takes a full response that might contain other strings and attempts to extract the JSON payload.
    Has support for truncated JSON where the JSON begins but the token window ends before the json is
    is properly closed.
    """
    # Deal with fully included responses as well as truncated responses that only have one
    if full_response.startswith("{") and not full_response.endswith("}"):
        full_response += "}"
    
    extracted_responses = list(_json_pattern.finditer(full_response))
    if not extracted_responses:
        logger.error(
            f"Unable to find any responses of the matching type `{full_response}`"
        )
        return None

    if len(extracted_responses) > 1 and verbose:
        logger.error(f"Unexpected response > 1, continuing anyway... {extracted_responses}")

    extracted_response = extracted_responses[0]

    if is_truncated(extracted_response.group(0)):
        # Start at the same location and just expand to the end of the message
        extracted_str = full_response[extracted_response.start() :]
    else:
        extracted_str = extracted_response.group(0)

    return extracted_str

def try_load_json(
    text: str,
    object_hook: Optional[Callable] = None,
    **kwargs,
):
    """
    Attempts to load the text as JSON
    """
    try:
        return json.loads(text, object_hook = object_hook, **kwargs)
    except Exception as e1:
        try:
            return json.loads(fix_json(text), object_hook = object_hook, **kwargs)
        except Exception as e2:
            logger.error(f"Unable to load JSON. Errors: {e1}, {e2}")
            raise e2


def extract_json_response(
    full_response: str, 
    verbose: Optional[bool] = False,
    raise_exceptions: Optional[bool] = False,
    object_hook: Optional[Callable] = None,
) -> Union[Dict[str, Any], List[Any], Any]:
    """
    Returns the extracted JSON response from the full response
    """
    extracted_str = find_json_response(full_response, verbose = verbose)
    if not extracted_str:
        return None
    try:
        return try_load_json(extracted_str, object_hook = object_hook)
    except Exception as e:
        if verbose: logger.trace(f"Unable to extract JSON response from {extracted_str}", error = e)
        if raise_exceptions: raise e
        return None


from __future__ import annotations

import functools
import tiktoken
import contextlib
from typing import Optional, Union, List, Dict, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from async_openai.schemas.chat import ChatMessage

def modelname_to_contextsize(modelname: str) -> int:
    """
    Calculate the maximum number of tokens possible to generate for a model.
    
    text-davinci-003: 4,097 tokens
    text-curie-001: 2,048 tokens
    text-babbage-001: 2,048 tokens
    text-ada-001: 2,048 tokens
    code-davinci-002: 8,000 tokens
    code-cushman-001: 2,048 tokens
    gpt-3.5-turbo: 4,096 tokens
    gpt-3.5-turbo-16k: 16,384 tokens
    gpt-4: 8,192 tokens
    gpt-4-32k: 32,768 tokens

    Args:
        modelname: The modelname we want to know the context size for.

    Returns:
        The maximum context size
    """
    if modelname == "code-davinci-002":
        return 8000

    if modelname in {
        "text-curie-001",
        "text-babbage-001",
        "text-ada-001",
        "code-cushman-001",
    }:
        return 2048

    # Check GPT4
    if modelname.startswith("gpt-4") or modelname.startswith("gpt4"):
        if "32k" in modelname:
            return 32768
        return 16384 if "16k" in modelname else 8192
    
    # Check GPT3.5
    if "gpt" in modelname \
        and "turbo" in modelname \
        and ("3.5" in modelname or "35" in modelname):
        return 16384 if "16k" in modelname else 4096
    
    return 4097
    

def get_encoder(
    model_name: str,
) -> tiktoken.Encoding:
    """
    Returns the correct encoder for the model name.
    """
    if "gpt" in model_name and "2" not in model_name:

        # Likely GPT4 or GPT3.5
        return tiktoken.get_encoding("cl100k_base")
    encoder = "gpt2"
    if model_name in {"text-davinci-003", "text-davinci-002"}:
        encoder = "p50k_base"
    if model_name.startswith("code"):
        encoder = "p50k_base"
    
    return tiktoken.get_encoding(encoder)

@functools.lru_cache(maxsize = 2048)
def get_token_count(
    text: str,
    model_name: str,
) -> int:
    """
    Returns the number of tokens in the text.
    """
    return len(get_encoder(model_name).encode(text))


def get_max_tokens(
    text: Union[str, List[str]],
    model_name: str,
    max_tokens: Optional[int] = None,
    padding_token_count: Optional[int] = 16 # tokens added to make sure we do not go over the limit
):
    """
    Returns the maximum number of tokens that can be generated for a model.
    """
    max_model_tokens = modelname_to_contextsize(model_name) - padding_token_count
    if isinstance(text, list):
        all_text_tokens = [get_token_count(t, model_name) for t in text]
        text_tokens = max(all_text_tokens)
    else:
        text_tokens = get_token_count(text, model_name)
    max_input_tokens = max_model_tokens - text_tokens
    if max_tokens is None:
        return max_input_tokens
    return min(max_input_tokens, max_tokens)
    # return modelname_to_contextsize(model_name) - get_token_count(text, model_name)


def get_chat_tokens_count(
    messages: List[Union[Dict[str, str], 'ChatMessage']],
    model_name: str,
    reply_padding_token_count: Optional[int] = 3,
    message_padding_token_count: Optional[int] = 4,
    **kwargs
) -> int:
    """
    Returns the number of tokens in the chat.
    """
    num_tokens = 0
    for message in messages:
        if message.get('name'):
            num_tokens -= 1
        num_tokens += message_padding_token_count + get_token_count(message.get('content', ''), model_name)

    num_tokens += reply_padding_token_count  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens

def get_max_chat_tokens(
    messages: List[Union[Dict[str, str], 'ChatMessage']],
    model_name: str,
    max_tokens: Optional[int] = None,
    reply_padding_token_count: Optional[int] = 3,
    message_padding_token_count: Optional[int] = 4,
    padding_token_count: Optional[int] = 16 # tokens added to make sure we do not go over the li
):
    """
    Returns the maximum number of tokens that can be generated for a model.
    """

    num_tokens = 0
    for message in messages:
        if message.get('name'):
            num_tokens -= 1
        num_tokens += message_padding_token_count + get_token_count(message.get('content', ''), model_name)

    num_tokens += reply_padding_token_count  # every reply is primed with <|start|>assistant<|message|>
    max_model_tokens = modelname_to_contextsize(model_name) - padding_token_count
    max_input_tokens = max_model_tokens - num_tokens
    if max_tokens is None:
        return max_input_tokens
    return min(max_input_tokens, max_tokens)


def fast_tokenize(text: Any) -> int:
    """
    Do a very fast tokenization of the text
    by estimating the number of tokens based on the
    number of characters in the string.

    1 token ~= 4 characters
    """
    return len(str(text)) // 4


from __future__ import annotations

"""
OpenAI Cost Functions and Handler
"""
import tiktoken
from pathlib import Path
from lazyops.types import BaseModel, validator, Field, lazyproperty
from typing import Optional, Union, Dict, Any, List, Tuple, Type, TYPE_CHECKING

if TYPE_CHECKING:
    from .resources import Usage
    from async_openai.schemas.chat import ChatMessage

pricing_file_path = Path(__file__).parent.joinpath('pricing.yaml')

class ModelCosts(BaseModel):
    """
    Represents a model's costs
    """
    unit: Optional[int] = 1000
    input: Optional[float] = 0.0
    output: Optional[float] = 0.0
    total: Optional[float] = 0.0


class ModelCostItem(BaseModel):
    """
    Represents a model's Cost Item
    """
    name: str
    aliases: Optional[List[str]] = None
    context_length: Optional[int] = 0
    costs: Optional[ModelCosts] = Field(default_factory=ModelCosts)
    endpoints: Optional[List[str]] = None

    def get_costs(
        self, 
        input_tokens: Optional[int] = None,
        output_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        usage: Optional['Usage'] = None,
        **kwargs
    ) -> float:
        """
        Gets the costs
        """
        if usage is not None:
            input_tokens = usage.prompt_tokens
            output_tokens = usage.completion_tokens
        if kwargs.get('prompt_tokens'):
            input_tokens = kwargs['prompt_tokens']
        if kwargs.get('completion_tokens'):
            output_tokens = kwargs['completion_tokens']
        
        assert input_tokens is not None or output_tokens is not None or total_tokens is not None, "Must provide either input_tokens, output_tokens, or total_tokens"
        if self.costs is None: return 0.0
        cost = 0.0
        if self.costs.input:
            cost += self.costs.input * input_tokens / self.costs.unit
        if self.costs.output:
            cost += self.costs.output * output_tokens / self.costs.unit
        if self.costs.total and total_tokens is not None:
            cost += self.costs.total * total_tokens / self.costs.unit
        return cost
        


class ModelContextHandlerMetaClass(type):
    """
    The Model Cost Handler
    """

    _models: Optional[Dict[str, ModelCostItem]] = None
    _model_aliases: Optional[Dict[str, str]] = None
    tokenizers: Optional[Dict[str, tiktoken.Encoding]] = {}

    def load_models(cls) -> Dict[str, ModelCostItem]:
        """
        Loads the models
        """
        import yaml
        models: Dict[str, Dict[str, Any]] = yaml.safe_load(pricing_file_path.read_text())
        return {k: ModelCostItem(name = k, **v) for k, v in models.items()}

    @property
    def models(cls) -> Dict[str, ModelCostItem]:
        """
        Gets the models
        """
        if cls._models is None:
            cls._models = cls.load_models()
        return cls._models

    @property
    def model_aliases(cls) -> Dict[str, str]:
        """
        Gets the model aliases
        """
        if cls._model_aliases is None:
            cls._model_aliases = {alias: model for model, item in cls.models.items() for alias in item.aliases or []}
        return cls._model_aliases
    
    def resolve_model_name(cls, model_name: str) -> str:
        """
        Resolves the Model Name from the model aliases
        """
        # Try to remove the version number
        key = model_name.rsplit('-', 1)[0].strip()
        if key in cls.model_aliases:
            cls.model_aliases[model_name] = cls.model_aliases[key]
        if key in cls.models:
            cls.model_aliases[model_name] = key
            return key
        raise KeyError(f"Model {model_name} not found")
    
    def __getitem__(cls, key: str) -> ModelCostItem:
        """
        Gets a model by name
        """
        if key not in cls.model_aliases and key not in cls.models:
            return cls.resolve_model_name(key)
        if key in cls.model_aliases:
            key = cls.model_aliases[key]
        return cls.models[key]
    
    def get(cls, key: str, default: Optional[str] = None) -> Optional[ModelCostItem]:
        """
        Gets a model by name
        """
        try:
            return cls[key]
        except KeyError:
            if default is None:
                raise KeyError(f"Model {key} not found") from None
            return cls[default]
        
    def add_model(cls, model: str, source_model: str):
        """
        Add a model to the handler
        
        Args:
            model (str): The model name
            source_model (str): The source model name
        """
        if model in cls.model_aliases or model in cls.models:
            return
        
        src_model = cls[source_model]
        # Add to the model aliases
        cls.model_aliases[model] = src_model.name


    def get_tokenizer(cls, name: str) -> Optional[tiktoken.Encoding]:
        """
        Gets the tokenizer
        """
        # Switch the 35 -> 3.5
        if '35' in name: name = name.replace('35', '3.5')
        if name not in cls.tokenizers:
            cls.tokenizers[name] = tiktoken.encoding_for_model(name)
        return cls.tokenizers[name]
    
    def count_chat_tokens(
        cls, 
        messages: List[Union[Dict[str, str], 'ChatMessage']],
        model_name: str,
        reply_padding_token_count: Optional[int] = 3,
        message_padding_token_count: Optional[int] = 4,
        **kwargs
    ) -> int:
        """
        Returns the number of tokens in the chat.
        """
        num_tokens = 0
        tokenizer = cls.get_tokenizer(model_name)
        for message in messages:
            if message.get('name'):
                num_tokens -= 1
            num_tokens += message_padding_token_count + len(tokenizer.encode(message.get('content', '')))
        num_tokens += reply_padding_token_count  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens

    def count_tokens(
        cls,
        text: Union[str, List[str]],
        model_name: str,
        **kwargs
    ) -> int:
        """
        Returns the number of tokens in the text.
        """
        tokenizer = cls.get_tokenizer(model_name)
        return (
            sum(len(tokenizer.encode(t)) for t in text)
            if isinstance(text, list)
            else len(tokenizer.encode(text))
        )
    
    def get_consumption_cost(cls, model_name: str, usage: 'Usage', **kwargs) -> float:
        """
        Gets the consumption cost
        """
        # Switch the 35 -> 3.5
        if '35' in model_name: model_name = model_name.replace('35', '3.5')
        model = cls[model_name]
        return model.get_costs(usage = usage, **kwargs)
    
    def resolve_model_name(cls, model_name: str) -> str:
        """
        Resolves the Model Name from the model aliases
        """
        return cls.model_aliases.get(model_name, model_name)
    
    def truncate_to_max_length(cls, text: str, model_name: str, context_length: Optional[int] = None, **kwargs) -> str:
        """
        Truncates the text to the max length
        """
        tokenizer = cls.get_tokenizer(model_name)
        if context_length is None:
            context_length = cls[model_name].context_length

        tokens = tokenizer.encode(text)
        if len(tokens) > context_length:
            tokens = tokens[-context_length:]
            decoded = tokenizer.decode(tokens)
            text = text[-len(decoded):]
        
        return text



class ModelContextHandler(metaclass = ModelContextHandlerMetaClass):
    """
    The Model Cost Handler
    """
    pass


import aiohttpx

from typing import Optional, Dict, Callable, List, TYPE_CHECKING
from async_openai.schemas import *
from async_openai.utils.config import get_settings, OpenAISettings, AzureOpenAISettings
from async_openai.utils.logs import logger


RouteClasses = {
    'completions': CompletionRoute,
    'chat': ChatRoute,
    'edits': EditRoute,
    'embeddings': EmbeddingRoute,
    # 'files': FileRoute,
    'images': ImageRoute,
    'models': ModelRoute,

}

class ApiRoutes:

    """
    Container for all the routes in the API.
    """

    completions: CompletionRoute = None
    chat: ChatRoute = None
    edits: EditRoute = None
    embeddings: EmbeddingRoute = None
    # files: FileRoute = None
    images: ImageRoute = None
    models: ModelRoute = None
    
    def __init__(
        self,
        client: aiohttpx.Client,
        name: str,
        # headers: Optional[Dict] = None,
        debug_enabled: Optional[bool] = False,
        on_error: Optional[Callable] = None,
        ignore_errors: Optional[bool] = False,
        disable_retries: Optional[bool] = None,
        retry_function: Optional[Callable] = None,

        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        settings: Optional[OpenAISettings] = None,
        is_azure: Optional[bool] = None,
        client_callbacks: Optional[List[Callable]] = None,
        
        **kwargs
    ):
        self.client = client
        self.name = name
        self.settings = settings or get_settings()
        # self.headers = headers or self.settings.get_headers()
        self.debug_enabled = debug_enabled
        self.on_error = on_error
        self.ignore_errors = ignore_errors
        self.disable_retries = disable_retries
        self.retry_function = retry_function

        self.timeout = timeout
        self.max_retries = max_retries
        self.is_azure = is_azure if is_azure is not None else \
            isinstance(self.settings, AzureOpenAISettings)
        self.kwargs = kwargs or {}
        if client_callbacks:
            self.kwargs['client_callbacks'] = client_callbacks
        self.init_routes()
    


    def init_routes(self):
        """
        Initializes the routes
        """
        for route, route_class in RouteClasses.items():
            try:
                setattr(self, route, route_class(
                    client = self.client,
                    name = self.name,
                    # headers = self.headers,
                    debug_enabled = self.debug_enabled,
                    on_error = self.on_error,
                    ignore_errors = self.ignore_errors,
                    disable_retries = self.disable_retries,
                    retry_function = self.retry_function,
                    timeout = self.timeout,
                    max_retries = self.max_retries,
                    settings = self.settings,
                    is_azure = self.is_azure,
                    **self.kwargs
                ))
            except Exception as e:
                logger.error(f"[{self.name}] Failed to initialize route {route} with error: {e}")
                raise e



import datetime
from typing import Optional, Type, List, Dict, Any
from lazyops.types import lazyproperty, Field

from async_openai.types.resources import BaseResource, Permission
from async_openai.types.responses import BaseResponse
from async_openai.types.routes import BaseRoute


__all__ = [
    'ModelData',
    'ModelObject',
    'ModelResponse',
    'ModelRoute',
]


class ModelData(BaseResource):
    id: str
    status: Optional[str] = None
    owned_by: Optional[str] = None
    created: Optional[datetime.datetime] = None
    updated_at: Optional[datetime.datetime] = None
    created_at: Optional[datetime.datetime] = None
    permission: Optional[List[Permission]] = Field(default_factory = list)
    root: Optional[str] = None
    parent: Optional[str] = None
    object: Optional[str] = 'model'
    capabilities: Optional[Dict[str, Any]] = Field(default_factory = dict)
    lifecycle_status: Optional[str] = None
    deprecation: Optional[Dict[str, Any]] = Field(default_factory = dict)


    @lazyproperty
    def model_age(self) -> Optional[datetime.datetime]:
        """
        Returns how long ago the model was created
        """
        if self.created:
            return datetime.datetime.now(tz = datetime.timezone.utc) - self.created


class ModelObject(BaseResource):
    model: Optional[str]


class ModelResponse(BaseResponse):
    data: Optional[List[ModelData]]
    data_model: Optional[Type[BaseResource]] = ModelData

    @lazyproperty
    def model_list(self) -> List[str]:
        return [model.id for model in self.data] if self.data and isinstance(self.data, list) else []


class ModelRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = ModelObject
    response_model: Optional[Type[BaseResource]] = ModelResponse

    @lazyproperty
    def api_resource(self):
        return 'models'
    
    @lazyproperty
    def create_enabled(self):
        return False
    
    @lazyproperty
    def list_enabled(self):
        return True
    
    @lazyproperty
    def get_enabled(self):
        return True

    def retrieve(
        self, 
        resource_id: str, 
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ModelResponse:
        """
        Retrieve a Single Model by Resource ID

        :param resource_id: The ID of the Resource to GET
        :param params: Optional Query Parameters
        """
        return super().retrieve(resource_id = resource_id, params = params, **kwargs)
    
    async def async_retrieve(
        self,
        resource_id: str,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ModelResponse:
        """
        Retrieve a Single Model by Resource ID

        :param resource_id: The ID of the Resource to GET
        :param param
        """
        return await super().async_retrieve(resource_id = resource_id, params = params, **kwargs)


    def list(
        self, 
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ModelResponse:
        """
        List all available Models

        :param params: Optional Query Parameters
        """
        return super().list(params = params, **kwargs)
    
    async def async_list(
        self, 
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ModelResponse:
        """
        List all available Models

        :param params: Optional Query Parameters
        """
        return await super().async_list(params = params, **kwargs)

    



